package be.iminds.ilabt.jfed.testing.base;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

/**
 * JsonTestConfigEditor
 */
public class JsonTestConfigEditor<C extends ApiTestConfig> implements ApiTestConfigEditor<C> {
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final Logger LOG = LoggerFactory.getLogger(JsonTestConfigEditor.class);

    @Nonnull private final Class<C> apiTestConfigClass;

    @Nonnull private Map curMap = new HashMap();

    public JsonTestConfigEditor(@Nonnull Class<C> apiTestConfigClass) {
        this.apiTestConfigClass = apiTestConfigClass;
    }

    @Nonnull
    @Override
    public List<ApiTestConfigParameter> getParameters() {
        List<ApiTestConfigParameter> parameters = new ArrayList<>();

        C defaultObject = null;
        try {
            defaultObject = apiTestConfigClass.newInstance();
        } catch (InstantiationException | IllegalAccessException e) {
            throw new RuntimeException("Reflection problem creating default ApiTestConfig. Probably the config is not in the expected/supported structure.", e);
        }

        addParameters(parameters, Collections.emptyList(), apiTestConfigClass, defaultObject);

        return parameters;
    }

    public void addParameters(List<ApiTestConfigParameter> parameters, List<String> nest, Class clazz, Object defaultObject) {
        LOG.debug("addParameters(_, "+nest+", "+clazz.getName()+", "+(defaultObject == null ? "null" : defaultObject.getClass().getName())+")");

        Method[] methods =
                defaultObject == null ?
                        clazz.getMethods():
                        defaultObject.getClass().getMethods();
        //only a very specific jackson annotation style is supported.
        for (Method method : methods) {
            if ((method.getName().startsWith("get") || method.getName().startsWith("is"))
                    && method.getParameters().length == 0) {
                JsonProperty jsonPropertAnnotation = method.getAnnotation(JsonProperty.class);
                JsonIgnore jsonIgnoreAnnotation = method.getAnnotation(JsonIgnore.class);
                ApiTestConfigProperty propertyAnnotation = method.getAnnotation(ApiTestConfigProperty.class);
                Nonnull nonnullAnnotation = method.getAnnotation(Nonnull.class);
                Nullable nullableAnnotation = method.getAnnotation(Nullable.class);
                boolean noEdit = propertyAnnotation != null && propertyAnnotation.noEdit();

                boolean isProperty = jsonPropertAnnotation != null;
                if (jsonIgnoreAnnotation != null && jsonIgnoreAnnotation.value()) isProperty = false;
                if (isProperty) {
                    String name = jsonPropertAnnotation.value();
                    int prefixLen = method.getName().startsWith("is") ? 2 : 3;
                    if (name.isEmpty() && method.getName().length() >= prefixLen+1) {
                        char first = Character.toLowerCase(method.getName().charAt(prefixLen));
                        name = first + method.getName().substring(prefixLen+1, method.getName().length());
                    }

                    //find the default by using the objects default contructor
                    Object defaultValue;
                    if (defaultObject != null) {
                        try {
                            defaultValue = method.invoke(defaultObject);
                        } catch (IllegalAccessException | InvocationTargetException e) {
                            throw new RuntimeException("Reflection problem traversing ApiTestConfig. Probably the config is not in the expected/supported structure. " +
                                    "defaultObject.class=" + (defaultObject == null ? "null" : defaultObject.getClass().getName()) +
                                    " clazz=" + clazz.getName(), e);
                        }
                    } else {
                        defaultValue = null;
                        //dirty hack: just discard it for now
                        return;
                    }

                    Class returnType = method.getReturnType();

                    assert returnType != null;
                    LOG.debug("Checking method "+method.getName()+
                            " returning "+returnType.getName()+
                            " in package "+(returnType.getPackage() == null ? null : returnType.getPackage().getName()));

                    boolean isEditableType = returnType.isPrimitive()
                            || returnType.getPackage() == null
                            || returnType.getPackage().getName().startsWith("java.lang")
                            || returnType.isEnum();

                    boolean isPojoReturn = !isEditableType;

                    boolean isCollectionReturn = !returnType.isPrimitive()
                            && returnType.getPackage() != null
                            && Collection.class.isAssignableFrom(returnType);

                    if (noEdit) {
                        //stop processing, so that no editor is created at all
                        isCollectionReturn = false;
                        isEditableType = false;
                        isPojoReturn = false;
                    }

                    List<String> newNest = new ArrayList<>(nest.size()+1);
                    newNest.addAll(nest);
                    if (!isCollectionReturn) {
                        if (isEditableType) {
                            newNest.add(name);
                            ApiTestConfigParameter param = new ApiTestConfigParameter(
                                    newNest,
                                    propertyAnnotation == null ? null : (propertyAnnotation.description().isEmpty() ? null : propertyAnnotation.description()),
                                    method.getReturnType(),
                                    defaultValue,
                                    nonnullAnnotation == null,
                                    propertyAnnotation == null ? null : (propertyAnnotation.editorHint().isEmpty() ? null : propertyAnnotation.editorHint()));
                            parameters.add(param);
                        } else {
                            if (isPojoReturn) {
                                newNest.add(name);
                                //recurse and use nested parameters  (take care not to make a new default when recursing, but to use defaultValue!)
                                addParameters(parameters, newNest, method.getReturnType(), defaultValue);
                            }
                        }
                    } else {
                        if (defaultValue != null) {
                            assert defaultValue instanceof Collection;
                            assert List.class.isAssignableFrom(returnType) : "Only list is currently supported";
                            assert defaultValue instanceof List;
                            List defaultValueList = (List) defaultValue;
                            if (!defaultValueList.isEmpty()) {
                                Object firstElement = defaultValueList.get(0);
                                Class collectionClass = firstElement.getClass(); //easiest way to get generic type (more complex if we didn't have the first element)

                                newNest.add(name + "[0]");
                                addParameters(parameters,
                                        Collections.singletonList(name + "[0]"),
                                        collectionClass,
                                        firstElement);
                            } else {
                                newNest.add(name);
                                LOG.debug("EMPTY LIST of default value for parameter "+newNest);
                            }
                        }
                    }
                }
            }
        }
    }

    @Override
    public void setParameter(@Nonnull ApiTestConfigParameter parameter, @Nullable Object value) {
        Map m = getSubMap(parameter);
        m.put(parameter.getLastName(), value);
    }

    @Nullable
    @Override
    public Object getParameter(@Nonnull ApiTestConfigParameter parameter) {
        Map m = getSubMap(parameter);
        return m.get(parameter.getLastName());
    }

    /** will add create sub maps as needed */
    @Nonnull
    private Map getSubMap(@Nonnull ApiTestConfigParameter parameter) {
        List<String> keys = parameter.getName();
        Map res = curMap;
        for (int i = 0; i < keys.size()-1; i++) {
            String key = keys.get(i);
            boolean isCollection = key.endsWith("[0]");
            if (isCollection)
                key = key.substring(0, key.length()-3);

            Object val = res.get(key);

            if (isCollection) {
                if (val == null) {
                    Map newMap = new HashMap();
                    List newList = Collections.singletonList(newMap);
                    res.put(key, newList);
                    res = newMap;
                } else {
                    if (val instanceof List) {
                        List valList = (List) val;
                        Object firstval = valList.get(0);
                        if (firstval instanceof Map) {
                            res = (Map) firstval;
                        } else {
                            throw new RuntimeException("Invalid list nesting! Something must be wrong with the parameters. keys=" + keys + " i=" + i + " val=" + val + " valclass=" + val.getClass().getName()+" firstval class = "+firstval.getClass().getName());
                        }
                    } else {
                        throw new RuntimeException("Invalid list nesting! Something must be wrong with the parameters. keys=" + keys + " i=" + i + " val=" + val + " valclass=" + val.getClass().getName());
                    }
                }
            } else {
                if (val == null) {
                    Map newMap = new HashMap();
                    res.put(key, newMap);
                    res = newMap;
                } else {
                    if (val instanceof Map) {
                        res = (Map) val;
                    } else {
                        throw new RuntimeException("Invalid nesting! Something must be wrong with the parameters. keys=" + keys + " i=" + i + " val=" + val + " valclass=" + val.getClass().getName());
                    }
                }
            }
        }
        return res;
    }

    @Override
    public void setConfig(@Nonnull C apiTestConfig) {
        curMap = MAPPER.convertValue(apiTestConfig, Map.class);
    }

    @Nonnull
    @Override
    public C build() {
        return MAPPER.convertValue(curMap, apiTestConfigClass);
    }
}
